@@ -16,13 +16,29 @@
 # Refer to the README and COPYING files for full details of the license.
 #
 
+import ctypes
 import logging
 import os
 import socket
 import subprocess
 import threading
 import time
+
 from OVirtAgentLogic import AgentLogicBase, DataRetriverBase
+from hooks import Hooks
+
+
+# avoid pep8 warnings
+def import_json():
+    try:
+        import json
+        return json
+    except ImportError:
+        import simplejson
+        return simplejson
+
+
+json = import_json()
 
 CredServer = None
 try:
@@ -37,6 +53,14 @@ except ImportError:
     CredServer = CredServerFake
 
 
+_GUEST_SCRIPTS_INSTALL_PATH = '/usr/share/ovirt-guest-agent'
+_GUEST_HOOKS_CONFIG_PATH = '/etc/ovirt-guest-agent/hooks.d'
+
+
+def _get_script_path(name):
+    return os.path.join(_GUEST_SCRIPTS_INSTALL_PATH, name)
+
+
 def _readLinesFromProcess(cmdline):
     try:
         process = subprocess.Popen(cmdline, stdout=subprocess.PIPE,
@@ -53,6 +77,35 @@ def _readLinesFromProcess(cmdline):
     return out.splitlines()
 
 
+class Container(object):
+    def __init__(self):
+        self.container = False
+        if 'container' in os.environ:
+            self.container = os.environ['container'] == 'docker'
+        if self.container:
+            self.libc = ctypes.CDLL('libc.so.6')
+            try:
+                self.selffd = os.open('/hostproc/self/ns/mnt', os.O_RDONLY)
+                self.hostfd = os.open('/hostproc/1/ns/mnt', os.O_RDONLY)
+            except (IOError, OSError):
+                # We can't leave anyway, so no need to even try
+                self.container = False
+                logging.warning('Failed to open mounts for container')
+
+    def resetns(self):
+        if self.container:
+            self.libc.setns(self.selffd, 0)
+
+    def setns(self):
+        if self.container:
+            self.libc.setns(self.hostfd, 0)
+
+    def make_procfs(self, path):
+        if self.container:
+            return path.replace('/proc/', '/hostproc/')
+        return path
+
+
 class PkgMgr(object):
 
     def rpm_list_packages(self, app_list):
@@ -148,19 +201,20 @@ class NicMgr(object):
             ipv6_addrs.append(ip.address)
         return ipv6_addrs
 
-    def ethtool_list_nics(self):
+    def ethtool_list_nics(self, blacklist=()):
         interfaces = list()
         try:
             for dev in self.ethtool.get_devices():
-                flags = self.ethtool.get_flags(dev)
-                if flags & self.ethtool.IFF_UP and \
-                        not(flags & self.ethtool.IFF_LOOPBACK):
-                    devinfo = self.ethtool.get_interfaces_info(dev)[0]
-                    interfaces.append(
-                        {'name': dev,
-                         'inet': self._get_ipv4_addresses(devinfo),
-                         'inet6': self._get_ipv6_addresses(devinfo),
-                         'hw': self.ethtool.get_hwaddr(dev)})
+                if (dev not in blacklist):
+                    flags = self.ethtool.get_flags(dev)
+                    if flags & self.ethtool.IFF_UP and \
+                            not(flags & self.ethtool.IFF_LOOPBACK):
+                        devinfo = self.ethtool.get_interfaces_info(dev)[0]
+                        interfaces.append(
+                            {'name': dev,
+                             'inet': self._get_ipv4_addresses(devinfo),
+                             'inet6': self._get_ipv6_addresses(devinfo),
+                             'hw': self.ethtool.get_hwaddr(dev)})
         except:
             logging.exception("Error retrieving network interfaces.")
         return interfaces
@@ -172,7 +226,7 @@ class CommandHandlerLinux:
         self.agent = agent
 
     def lock_screen(self):
-        cmd = ['/usr/share/ovirt-guest-agent/ovirt-locksession']
+        cmd = [_get_script_path('ovirt-locksession')]
         logging.debug("Executing lock session command: '%s'", cmd)
         subprocess.call(cmd)
 
@@ -180,7 +234,9 @@ class CommandHandlerLinux:
         self.agent.cred_server.user_authenticated(credentials)
 
     def logoff(self):
-        pass
+        CMD = ['/usr/share/ovirt-guest-agent/ovirt-logout']
+        logging.debug("Executing logout command: '%s'", CMD)
+        subprocess.call(CMD)
 
     def shutdown(self, timeout, msg, reboot=False):
         # The shutdown command works with minutes while vdsm send value in
@@ -191,14 +247,14 @@ class CommandHandlerLinux:
         if reboot:
             param = '-r'
             action = 'reboot'
-        cmd = ['/usr/share/ovirt-guest-agent/ovirt-shutdown', param,
+        cmd = [_get_script_path('ovirt-shutdown'), param,
                "+%d" % (delay), "\"%s\"" % (msg)]
 
         logging.debug("Executing %s command: %s", action, cmd)
         subprocess.call(cmd)
 
     def hibernate(self, state):
-        cmd = ['/usr/share/ovirt-guest-agent/ovirt-hibernate', state]
+        cmd = [_get_script_path('ovirt-hibernate'), state]
         logging.debug("Executing hibernate command: %s", cmd)
         subprocess.call(cmd)
 
@@ -209,6 +265,7 @@ class CommandHandlerLinux:
 class LinuxDataRetriver(DataRetriverBase):
 
     def __init__(self):
+        self.container = Container()
         try:
             pkgmgr = PkgMgr()
         except NotImplementedError:
@@ -218,11 +275,12 @@ class LinuxDataRetriver(DataRetriverBase):
         try:
             nicmgr = NicMgr()
         except NotImplementedError:
-            self.list_nics = lambda: []
+            self.list_nics = lambda _: []
         else:
             self.list_nics = nicmgr.list_nics
         self.app_list = ""
         self.ignored_fs = ""
+        self.ignored_nics = ""
         self.ignore_zero_size_fs = True
         self._init_vmstat()
         DataRetriverBase.__init__(self)
@@ -233,45 +291,85 @@ class LinuxDataRetriver(DataRetriverBase):
     def getOsVersion(self):
         return os.uname()[2]
 
+    def getContainerList(self):
+        cmd = [_get_script_path('ovirt-container-list')]
+        # skip if not available
+        if not os.path.exists(cmd[0]):
+            return []
+        logging.debug('Executing ovirt-container-list command')
+        result = []
+        try:
+            p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+            result = json.loads(p.stdout.read())
+        except Exception:
+            logging.exception('ovirt-container-list invocation failed')
+        return result
+
+    def getOsInfo(self):
+        cmd = [_get_script_path('ovirt-osinfo')]
+        logging.debug('Executing ovirt-osinfo command: %s', cmd)
+        result = {
+            'version': '',
+            'distribution': '',
+            'codename': '',
+            'arch': '',
+            'type': 'linux',
+            'kernel': ''}
+        try:
+            p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
+            for line in p.stdout.read().split('\n'):
+                line = line.strip()
+                if line:
+                    k, v = line.split('=', 1)
+                    if v and k in result:
+                        result[k] = v
+        except Exception:
+            logging.exception('ovirt-osinfo invocation failed')
+        return result
+
     def getAllNetworkInterfaces(self):
-        return self.list_nics()
+        return self.list_nics(self.ignored_nics.split())
 
     def getApplications(self):
         return self.list_pkgs(self.app_list)
 
     def getAvailableRAM(self):
         free = 0
-        for line in open('/proc/meminfo'):
+        for line in open(self.container.make_procfs('/proc/meminfo')):
             var, value = line.strip().split()[0:2]
             if var in ('MemFree:', 'Buffers:', 'Cached:'):
                 free += long(value)
         return str(free / 1024)
 
     def getUsers(self):
+        self.container.setns()
         users = ''
         try:
             cmdline = '/usr/bin/users | /usr/bin/tr " " "\n" | /usr/bin/uniq'
             users = ' '.join(os.popen(cmdline).read().split())
         except:
             logging.exception("Error retrieving logged in users.")
+        self.container.resetns()
         return users
 
     def getActiveUser(self):
+        self.container.setns()
         users = os.popen('/usr/bin/users').read().split()
         try:
             user = users[0]
         except:
             user = 'None'
+        self.container.resetns()
         return user
 
     def getDisksUsage(self):
         usages = list()
         try:
-            mounts = open('/proc/mounts', 'r')
+            mounts = open(self.container.make_procfs('/proc/mounts'), 'r')
             for mount in mounts:
                 try:
                     (device, path, fs) = mount.split()[:3]
-                    if fs not in self.ignored_fs:
+                    if fs not in self.ignored_fs and not os.path.isfile(path):
                         # path might include spaces.
                         path = path.decode("string-escape")
                         statvfs = os.statvfs(path)
@@ -282,7 +380,8 @@ class LinuxDataRetriver(DataRetriverBase):
                         usages.append({'path': path, 'fs': fs, 'total': total,
                                        'used': used})
                 except:
-                    logging.exception("Error retrieving disks usages.")
+                    logging.exception("Error retrieving disk usage; "
+                                      "mount description: %s", mount)
             mounts.close()
         except Exception:
             logging.exception("Error during reading mounted devices")
@@ -324,7 +423,7 @@ class LinuxDataRetriver(DataRetriverBase):
         fields = {'MemTotal:': 0, 'MemFree:': 0, 'Buffers:': 0,
                   'Cached:': 0, 'SwapFree:': 0, 'SwapTotal:': 0}
         free = 0
-        for line in open('/proc/meminfo'):
+        for line in open(self.container.make_procfs('/proc/meminfo')):
             (key, value) = line.strip().split()[0:2]
             if key in fields.keys():
                 fields[key] = int(value)
@@ -352,7 +451,7 @@ class LinuxDataRetriver(DataRetriverBase):
         interval = self.vmstat['timestamp_cur'] - self.vmstat['timestamp_prev']
         self.vmstat['timestamp_prev'] = self.vmstat['timestamp_cur']
 
-        for line in open('/proc/vmstat'):
+        for line in open(self.container.make_procfs('/proc/vmstat')):
             (key, value) = line.strip().split()[0:2]
             if key in fields.keys():
                 name = fields[key]
@@ -371,11 +470,14 @@ class LinuxVdsAgent(AgentLogicBase):
         AgentLogicBase.__init__(self, config)
         self.dr = LinuxDataRetriver()
         self.dr.app_list = config.get("general", "applications_list")
+        self.dr.ignored_nics = config.get("general", "ignored_nics")
         self.dr.ignored_fs = set(config.get("general", "ignored_fs").split())
         self.dr.ignore_zero_size_fs = config.get("general",
                                                  "ignore_zero_size_fs")
         self.commandHandler = CommandHandlerLinux(self)
         self.cred_server = CredServer()
+        self.hooks = Hooks(logging.getLogger('Hooks'),
+                           _GUEST_HOOKS_CONFIG_PATH)
 
     def run(self):
         self.cred_server.start()
@@ -390,6 +492,7 @@ def test():
     from pprint import pprint
     dr = LinuxDataRetriver()
     dr.app_list = "kernel kernel-headers aspell"
+    dr.ignored_nics = "docker0"
     dr.ignored_fs = set("rootfs tmpfs autofs cgroup selinuxfs udev mqueue "
                         "nfsd proc sysfs devtmpfs hugetlbfs rpc_pipefs devpts "
                         "securityfs debugfs binfmt_misc fuse.gvfsd-fuse "
@@ -410,5 +513,6 @@ def test():
     pprint(dr.getDiskMapping())
     print "Memory Stats:", dr.getMemoryStats()
 
+
 if __name__ == '__main__':
     test()
