@@ -1,26 +1,32 @@
-#!/usr/bin/python
+#!/usr/bin/python2
 
-import time
-import os
-import logging
-import win32netcon
-import win32net
-import win32ts
-import win32api
-import win32pipe
-import win32security
-import win32process
-import win32con
-import win32com.client
-import pythoncom
-import subprocess
-import socket
-from OVirtAgentLogic import AgentLogicBase, DataRetriverBase
+import _winreg
+import ctypes
 from ctypes import c_ulong, byref, windll, create_unicode_buffer,\
     Structure, sizeof, c_void_p
 from ctypes.util import find_library
 from ctypes.wintypes import DWORD
-import _winreg
+import ctypes.wintypes
+import logging
+import os
+import socket
+import subprocess
+import time
+
+import pythoncom
+import win32api
+import win32com.client
+import win32con
+import win32file
+import win32net
+import win32netcon
+import win32pipe
+import win32process
+import win32security
+import win32ts
+
+from OVirtAgentLogic import AgentLogicBase, DataRetriverBase
+from hooks import Hooks
 
 
 # Constants according to
@@ -29,12 +35,38 @@ KEY_WOW64_32KEY = 0x0100
 KEY_WOW64_64KEY = 0x0200
 
 
+class StoragePropertyQuery(ctypes.Structure):
+    _fields_ = (
+        ("a", ctypes.wintypes.ULONG),
+        ("b", ctypes.wintypes.ULONG),
+        ("c", ctypes.wintypes.BYTE),
+    )
+
+
+class StorageDeviceDescriptor(ctypes.Structure):
+    _fields_ = (
+        ("Version", ctypes.wintypes.ULONG),
+        ("Size", ctypes.wintypes.ULONG),
+        ("DeviceType", ctypes.wintypes.BYTE),
+        ("DeviceTypeModifier", ctypes.wintypes.BYTE),
+        ("RemovableMedia", ctypes.wintypes.BOOLEAN),
+        ("CommandQueueing", ctypes.wintypes.BOOLEAN),
+        ("VendorIdOffset", ctypes.wintypes.ULONG),
+        ("ProductIdOffset", ctypes.wintypes.ULONG),
+        ("ProductRevisionOffset", ctypes.wintypes.ULONG),
+        ("SerialNumberOffset", ctypes.wintypes.ULONG),
+        ("BusType", ctypes.wintypes.DWORD),
+        ("RawPropertiesLength", ctypes.wintypes.ULONG),
+        ("DATA", ctypes.wintypes.BYTE * 1024),
+    )
+
+
 # _winreg.QueryValueEx and win32api.RegQueryValueEx don't support reading
 # Unicode strings from the registry (at least on Python 2.5.1).
 def QueryStringValue(hkey, name):
     # if type(hkey) != type(PyHKEY):
     #     raise TypeError("1nd arg must be a PyHKEY.")
-    if type(name) != type(unicode()):
+    if not isinstance(name, unicode):
         raise TypeError("2nd arg must be a unicode.")
     key_type = c_ulong(0)
     key_len = c_ulong(0)
@@ -136,63 +168,106 @@ class IncomingMessageTypes:
         Credentials = 11
 
 
+class OSVERSIONINFOEXW(ctypes.Structure):
+    _fields_ = [('dwOSVersionInfoSize', ctypes.c_ulong),
+                ('dwMajorVersion', ctypes.c_ulong),
+                ('dwMinorVersion', ctypes.c_ulong),
+                ('dwBuildNumber', ctypes.c_ulong),
+                ('dwPlatformId', ctypes.c_ulong),
+                ('szCSDVersion', ctypes.c_wchar * 128),
+                ('wServicePackMajor', ctypes.c_ushort),
+                ('wServicePackMinor', ctypes.c_ushort),
+                ('wSuiteMask', ctypes.c_ushort),
+                ('wProductType', ctypes.c_byte),
+                ('wReserved', ctypes.c_byte)]
+
+
 class WinOsTypeHandler:
-    WINNT3_51 = 'Win NT 3.51'
-    WINNT4 = 'Win NT 4'
-    WIN2K = 'Win 2000'
-    WINXP = 'Win XP'
     WIN2003 = 'Win 2003'
     WIN2008 = 'Win 2008'
     WIN2008R2 = 'Win 2008 R2'
     WIN2012 = 'Win 2012'
+    WIN2012R2 = 'Win 2012 R2'
+    WIN2016 = 'Win 2016'
+    WIN2019 = 'Win 2019'
+    WIN2K = 'Win 2000'
+    WINXP = 'Win XP'
+    WINVISTA = 'Win Vista'
     WIN7 = 'Win 7'
     WIN8 = 'Win 8'
-    WINCE3_1_0 = 'Win CE 1.0'
-    WINCE3_2_0 = 'Win CE 2.0'
-    WINCE3_2_1 = 'Win CE 2.1'
-    WINCE3_3_0 = 'Win CE 3.0'
+    WIN81 = 'Win 8.1'
+    WIN10 = 'Win 10'
     UNKNOWN = 'Unknown'
-    # winVersionMatrix is constructed from 3 fields
-    # <platformId>.<MajorVersion>.<MinorVersion>
-    winVersionMatrix = {
-        '2.3.51': WINNT3_51,
-        '2.4.0': WINNT4,
-        '2.5.0': WIN2K,
-        '2.5.1': WINXP,
-        '2.5.2': WIN2003,
-        '2.6.0': WIN2008,
-        '2.6.1': WIN2008R2,  # Window Server 2008 R2
-        '2.6.2': WIN2012,
-        '3.1.0': WINCE3_1_0,
-        '3.2.0': WINCE3_2_0,
-        '3.2.1': WINCE3_2_1,
-        '3.3.0': WINCE3_3_0}
+
+    desktopVersionMatrix = {
+        '5.0': WIN2K,
+        '5.1': WINXP,
+        '6.0': WINVISTA,
+        '6.1': WIN7,
+        '6.2': WIN8,
+        '6.3': WIN81,
+        '10.0': WIN10,
+    }
+    serverVersionMatrix = {
+        '5.2': WIN2003,
+        '6.0': WIN2008,
+        '6.1': WIN2008R2,
+        '6.2': WIN2012,
+        '6.3': WIN2012R2,
+        '10.0': [
+            (17763, WIN2019),
+            (14393, WIN2016),
+        ],
+    }
 
     def getWinOsType(self):
-        retval = self.UNKNOWN
-        try:
-            versionTupple = win32api.GetVersionEx(1)
-            key = "%d.%d.%d" % (
-                versionTupple[3], versionTupple[0], versionTupple[1])
-            if key in self.winVersionMatrix:
-                retval = self.winVersionMatrix[key]
-            # Window 7 and Window Server 2008 R2 share the same version.
-            # Need to fix it using the wProductType field.
-                VER_NT_WORKSTATION = 1
-            if (retval == WinOsTypeHandler.WIN2008R2 and
-                    versionTupple[8] == VER_NT_WORKSTATION):
-                retval = WinOsTypeHandler.WIN7
-            elif (retval == WinOsTypeHandler.WIN2012 and
-                    versionTupple[8] == VER_NT_WORKSTATION):
-                retval = WinOsTypeHandler.WIN8
-            logging.debug("WinOsTypeHandler::getWinOsType osType = '%s'",
-                          retval)
-        except:
-            logging.exception("getWinOsType - failed")
-        return retval
+        name = self.UNKNOWN
+        version = ''
+        os_version = OSVERSIONINFOEXW()
+        os_version.dwOSVersionInfoSize = ctypes.sizeof(os_version)
+        retcode = ctypes.windll.Ntdll.RtlGetVersion(ctypes.byref(os_version))
+        VER_NT_WORKSTATION = 1
+        if retcode == 0:
+            major = os_version.dwMajorVersion
+            minor = os_version.dwMinorVersion
+            matrix = self.serverVersionMatrix
+            if os_version.wProductType == VER_NT_WORKSTATION:
+                matrix = self.desktopVersionMatrix
+            version = '%d.%d' % (major, minor)
+            name = matrix.get(version, '')
+            if isinstance(name, list):
+                build_list = name
+                name = ''
+                for final_build, build_name in build_list:
+                    if os_version.dwBuildNumber <= final_build:
+                        name = build_name
+        return {'name': name, 'version': version}
+
+
+def remove_bcd_useplatformclock():
+    try:
+        subprocess.call(['%SystemRoot%\\sysnative\\bcdedit.exe',
+                         '/deletevalue', '{current}', 'USEPLATFORMCLOCK'],
+                        shell=True)
+    except OSError:
+        logging.info('Failed to set the USEPLATFORMCLOCK flag via '
+                     'bcdedit.exe', exc_info=True)
+
+
+def apply_clock_tuning():
+    osinfo = WinOsTypeHandler().getWinOsType()
+    parts = osinfo.get('version', '6.0').split('.')
+    if parts[0] >= '6':
+        if parts[0] == '6' and parts[1] == '0':
+            # Skip 6.0 and lower
+            pass
+        else:
+            remove_bcd_useplatformclock()
 
 
 class CommandHandlerWin:
+    def __init__(self, dr):
+        self._dr = dr
 
     def lock_screen(self):
         self.LockWorkStation()
@@ -202,7 +277,7 @@ class CommandHandlerWin:
                    "\\Policies\\System"
         if value is not None:
             view_flag = KEY_WOW64_64KEY
-            handle = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, KEY_PATH,
+            handle = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, KEY_PATH, 0,
                                      view_flag | _winreg.KEY_READ |
                                      _winreg.KEY_WRITE)
             try:
@@ -210,8 +285,8 @@ class CommandHandlerWin:
             except OSError:
                 # Expected to happen if it does not exist
                 old = (None, None)
-            _winreg.SetValue(handle, 'SoftwareSASGeneration',
-                             _winreg.REG_DWORD, value)
+            _winreg.SetValueEx(handle, 'SoftwareSASGeneration', 0,
+                               _winreg.REG_DWORD, value)
             return old[0]
         return None
 
@@ -226,11 +301,17 @@ class CommandHandlerWin:
             self._setSoftwareSASPolicy(oldValue)
 
     def login(self, credentials):
+        if not self._dr.desktopLocked:
+            logging.debug("Not unlocking screen because it is not locked")
+            return
         PIPE_NAME = "\\\\.\\pipe\\VDSMDPipe"
         BUFSIZE = 1024
         RETIRES = 3
         try:
             self._performSAS()
+        except Exception:
+            logging.warning("Failed to perform SAS", exc_info=True)
+        try:
             retries = 1
             while retries <= RETIRES:
                 try:
@@ -355,6 +436,7 @@ class CommandHandlerWin:
 
 class WinDataRetriver(DataRetriverBase):
     def __init__(self):
+        self.arch = self._getArch()
         self.os = WinOsTypeHandler().getWinOsType()
         DataRetriverBase.__init__(self)
 
@@ -362,13 +444,41 @@ class WinDataRetriver(DataRetriverBase):
         return os.environ.get('COMPUTERNAME', '')
 
     def getOsVersion(self):
-        return self.os
+        return self.os['name']
+
+    def _getArch(self):
+        arch = 'x86'
+        try:
+            kernel32 = ctypes.windll.kernel32
+            result = ctypes.c_int()
+            proc = kernel32.GetCurrentProcess()
+            if kernel32.IsWow64Process(proc, ctypes.byref(result)) == 1:
+                if result:
+                    arch = 'x86_64'
+        except AttributeError:
+            pass
+        return arch
+
+    def getOsInfo(self):
+        return {
+            'version': self.os['version'],
+            'distribution': '',
+            'codename': self.os['name'],
+            'arch': self.arch,
+            'type': 'windows',
+            'kernel': ''}
+
+    def getContainerList(self):
+        return []
 
     def getAllNetworkInterfaces(self):
         return GetNetworkInterfaces()
 
     def _is_item_update(self, reg_key):
         RTPATTERNS = ("Hotfix", "Security Update", "Software Update", "Update")
+        help_link = QueryStringValue(reg_key, u'HelpLink')
+        if help_link.startswith('http://support.microsoft.com'):
+            return True
         release_type = QueryStringValue(reg_key, u'ReleaseType')
         for pattern in RTPATTERNS:
             if release_type.find(pattern) >= 0:
@@ -382,12 +492,13 @@ class WinDataRetriver(DataRetriverBase):
         retval = set()
         key_path = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall"
         for view_flag in (KEY_WOW64_32KEY, KEY_WOW64_64KEY):
-            rootkey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, key_path,
+            rootkey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, key_path, 0,
                                       view_flag | _winreg.KEY_READ)
             items = _winreg.QueryInfoKey(rootkey)[0]
             for idx in range(items):
                 cur_key_path = _winreg.EnumKey(rootkey, idx)
-                cur_key = _winreg.OpenKey(rootkey, cur_key_path)
+                cur_key = _winreg.OpenKey(rootkey, cur_key_path, 0,
+                                          view_flag | _winreg.KEY_READ)
                 try:
                     if self._is_item_update(cur_key):
                         continue
@@ -443,7 +554,8 @@ class WinDataRetriver(DataRetriverBase):
             elif domain.lower() != self.getMachineName().lower():
                 # Use FQDN as user name if computer is part of a domain.
                 try:
-                    user = u"%s\\%s" % (domain, user)
+                    user_orig = user
+                    user = u"%s\\%s" % (domain, user_orig)
                     user = win32security.TranslateName(
                         user,
                         win32api.NameSamCompatible,
@@ -454,7 +566,9 @@ class WinDataRetriver(DataRetriverBase):
                     if err != 0:
                         raise RuntimeError(err, 'TranslateName')
                 except:
-                    logging.exception("Error on user name translation.")
+                    logging.debug("Error on user name translation. Requested "
+                                  "translation '%s' '%s'", user_orig, domain)
+                    user = u"%s@%s" % (user_orig, domain)
             else:
                 user = u"%s@%s" % (user, domain)
         except:
@@ -488,6 +602,35 @@ class WinDataRetriver(DataRetriverBase):
             logging.exception("Error retrieving disks usages.")
         return usages
 
+    def _readSMART(self, name):
+        serial = 'NO SERIAL NUMBER'
+        handle = ctypes.windll.kernel32.CreateFileW(
+            name,
+            win32file.GENERIC_READ | win32file.GENERIC_WRITE,
+            win32file.FILE_SHARE_READ | win32file.FILE_SHARE_WRITE,
+            None, win32file.OPEN_EXISTING, 0, 0)
+        if handle == -1:
+            logging.warning("Failed to open device '%s' for querying the "
+                            "serial number. Error code: %d", name,
+                            win32api.GetLastError())
+            return serial
+        q = StoragePropertyQuery()
+        r = StorageDeviceDescriptor()
+        read_count = ctypes.wintypes.ULONG()
+        ret = ctypes.windll.kernel32.DeviceIoControl(
+            handle, 0x002D1400, ctypes.addressof(q), ctypes.sizeof(q),
+            ctypes.addressof(r), ctypes.sizeof(r),
+            ctypes.addressof(read_count), 0)
+
+        if ret:
+            serial = buffer(r)[r.SerialNumberOffset:r.SerialNumberOffset + 20]
+        else:
+            logging.warning("DeviceIoControl for device %s failed with"
+                            "eror code: %d. Could not look up serial number",
+                            name, win32api.GetLastError())
+        ctypes.windll.kernel32.CloseHandle(handle)
+        return serial
+
     def getDiskMapping(self):
         result = {}
         try:
@@ -500,7 +643,11 @@ class WinDataRetriver(DataRetriverBase):
                 objSWbemServices.ExecQuery(
                     "SELECT * FROM Win32_DiskDrive")
             for objItem in colItems:
-                result[objItem.SerialNumber] = {'name': objItem.DeviceID}
+                try:
+                    serial = objItem.SerialNumber
+                except AttributeError:
+                    serial = self._readSMART(objItem.DeviceID)
+                result[serial] = {'name': objItem.DeviceID}
         except Exception:
             logging.exception("Failed to retrieve disk mapping")
         return result
@@ -558,10 +705,15 @@ class WinDataRetriver(DataRetriverBase):
 
 class WinVdsAgent(AgentLogicBase):
 
-    def __init__(self, config):
+    def __init__(self, config, install_dir):
         AgentLogicBase.__init__(self, config)
         self.dr = WinDataRetriver()
-        self.commandHandler = CommandHandlerWin()
+        self.commandHandler = CommandHandlerWin(self.dr)
+        hooks_dir = os.path.join(install_dir, 'hooks')
+        self.hooks = Hooks(logging.getLogger('Hooks'), hooks_dir)
+
+        if config.getboolean('general', 'apply_timer_configuration'):
+            apply_clock_tuning()
 
     def run(self):
         logging.debug("WinVdsAgent:: run() entered")
@@ -571,10 +723,16 @@ class WinVdsAgent(AgentLogicBase):
         except:
             logging.exception("WinVdsAgent::run")
 
+    def doListen(self):
+        # CoInitializeEx() should be called in multi-threading program
+        # according to msdn document.
+        pythoncom.CoInitializeEx(0)
+        AgentLogicBase.doListen(self)
+
     def doWork(self):
-        # CoInitialize() should be called in multi-threading program according
-        # to msdn document.
-        pythoncom.CoInitialize()
+        # CoInitializeEx() should be called in multi-threading program
+        # according to msdn document.
+        pythoncom.CoInitializeEx(0)
         AgentLogicBase.doWork(self)
 
     def disable_screen_saver(self):
@@ -600,6 +758,8 @@ def test():
     print "Active User:", dr.getActiveUser()
     print "Disks Usage:", dr.getDisksUsage()
     print "Memory Stats:", dr.getMemoryStats()
+    print "DiskMapping:", dr.getDiskMapping()
+
 
 if __name__ == '__main__':
     test()
